{-# LANGUAGE GADTs, KindSignatures, FlexibleInstances, StandaloneDeriving, DeriveDataTypeable #-}
{-# OPTIONS -Wall #-}
module Language.Barracuda.AST where

------------------------------------------------------------------------
-- Imports
------------------------------------------------------------------------
-- for implementing instances for functions
import qualified Util.Unique as Unique

-- import Data.Data (Data)
import Data.Typeable (Typeable, Typeable1, Typeable2, Typeable3, cast)

------------------------------------------------------------------------
-- Scalar Barracuda Types
------------------------------------------------------------------------
-- some shorthand
type B = Bool
type F = Float
type I = Int

-- | Witnesses of the scalar types in Barracuda.
data ScalarWitness :: * -> * where
  WB :: ScalarWitness B
  WF :: ScalarWitness F
  WI :: ScalarWitness I

-- | Existential wrapper around scalar type witnesses.
data SomeScalarWitness :: * where
  SomeScalarWitness :: ScalarWitness a -> SomeScalarWitness

-- | A class for the scalar types in Barracuda.
class (Eq a, Ord a, Show a, Typeable a) => Scalar a where
  -- | Get a witness of the scalar's type.  The argument is not
  -- inspected.
  scalarWitness :: a -> ScalarWitness a

  -- | Convert a Haskell value into a scalar Barracuda expression.
  toScalar      :: a -> SExpr a

instance Scalar Float where
  scalarWitness _ = WF
  toScalar = SConst . SCF

instance Scalar Int where
  scalarWitness _ = WI
  toScalar = SConst . SCI

instance Scalar Bool where
  scalarWitness _ = WB
  toScalar = SConst . SCB

-- | Witness a scalar expression's type.
sexprType :: (Scalar a) => SExpr a -> ScalarWitness a
sexprType _ = scalarWitness undefined

-- | Witness an array expression's element type.
aexprElemType :: (Scalar a) => AExpr a -> ScalarWitness a
aexprElemType _ = scalarWitness undefined

-- | Witness a matrix expression's element type.
mexprElemType :: (Scalar a) => MExpr a -> ScalarWitness a
mexprElemType _ = scalarWitness undefined

------------------------------------------------------------------------
-- AST types
------------------------------------------------------------------------
-- | An identifier is simply a string.
type Id = String

-- | Scalar expressions.
data SExpr :: * -> * where
  SLet    :: (Scalar a, Scalar b) => SExpr a -> (SExpr a -> SExpr b) -> SExpr b
  SVar    :: (Scalar a) => Id -> SExpr a
  SConst  :: (Scalar a) => SConst a -> SExpr a
  Prim1   :: (Scalar a, Scalar b)
          => PrimOp1 a b
          -> SExpr a
          -> SExpr b
  Prim2   :: (Scalar a, Scalar b, Scalar c)
          => PrimOp2 a b c
          -> SExpr a
          -> SExpr b
          -> SExpr c
  SCond   :: (Scalar a) => SExpr B -> SExpr a -> SExpr a -> SExpr a
  AReduce :: (Scalar a)
          => (SExpr a -> SExpr a -> SExpr a)
          -> SExpr a
          -> AExpr a
          -> SExpr a
  AIdx    :: (Scalar a) => Id -> SExpr I -> SExpr a
  ALen    :: (Scalar a) => AExpr a -> SExpr I
  MReduce :: (Scalar a)
          => (SExpr a -> SExpr a -> SExpr a)
          -> SExpr a
          -> MExpr a
          -> SExpr a
  MIdx    :: (Scalar a)
          => Id           -- matrix variable to index
          -> SExpr I      -- row index
          -> SExpr I      -- column index
          -> SExpr a
  MRows   :: (Scalar a) => MExpr a -> SExpr I
  MCols   :: (Scalar a) => MExpr a -> SExpr I

-- | Array expressions.
data AExpr :: * -> * where
  AVar   :: (Scalar a)
         => Id
         -> AExpr a
  ASliceVar :: (Scalar a) => ASliceInfo -> Id -> AExpr a
  AMap   :: (Scalar a, Scalar b)
         => (SExpr a -> SExpr b)
         -> AExpr a
         -> AExpr b
  AMap2  :: (Scalar a, Scalar b, Scalar c)
         => (SExpr a -> SExpr b -> SExpr c)
         -> AExpr a
         -> AExpr b
         -> AExpr c
  AMap3  :: (Scalar a, Scalar b, Scalar c, Scalar d)
         => (SExpr a -> SExpr b -> SExpr c -> SExpr d)
         -> AExpr a
         -> AExpr b
         -> AExpr c
         -> AExpr d

-- | Matrix expressions.
data MExpr :: * -> * where
  MVar :: (Scalar a) => Id -> MExpr a
  MSliceVar :: (Scalar a) => MSliceInfo -> Id -> MExpr a
  MMap   :: (Scalar a, Scalar b)
         => (SExpr a -> SExpr b)
         -> MExpr a
         -> MExpr b
  MMap2  :: (Scalar a, Scalar b, Scalar c)
         => (SExpr a -> SExpr b -> SExpr c)
         -> MExpr a
         -> MExpr b
         -> MExpr c

data Slice :: * where
  Slice :: SExpr Int -> SExpr Int -> SExpr Int -> Slice

type ASliceInfo = Slice

data MSliceInfo = MSliceInfo { rowSlice :: Slice
                             , colSlice :: Slice
                             } deriving (Eq)

-- | Unary primitive scalar operations.  These are parameterized by
-- the type of the argument and the return type.
data PrimOp1 :: * -> * -> * where
  Not    :: PrimOp1 B B
  FNeg   :: PrimOp1 F F
  INeg   :: PrimOp1 I I
  FAbs   :: PrimOp1 F F
  IAbs   :: PrimOp1 I I
  I2F    :: PrimOp1 I F
  FSqrt  :: PrimOp1 F F
  FLog   :: PrimOp1 F F
  FExp   :: PrimOp1 F F
  FSin   :: PrimOp1 F F
  FCos   :: PrimOp1 F F
  FTan   :: PrimOp1 F F
  FASin  :: PrimOp1 F F
  FACos  :: PrimOp1 F F
  FATan  :: PrimOp1 F F
  FSinh  :: PrimOp1 F F
  FCosh  :: PrimOp1 F F
  FTanh  :: PrimOp1 F F
  FASinh :: PrimOp1 F F
  FACosh :: PrimOp1 F F
  FATanh :: PrimOp1 F F

-- | Binary primitive scalar operations.  These are parameterized by
-- the types of the arguments and the return type.
data PrimOp2 :: * -> * -> * -> * where
  FAdd :: PrimOp2 F F F
  FSub :: PrimOp2 F F F
  FMul :: PrimOp2 F F F
  FDiv :: PrimOp2 F F F

  FEq  :: PrimOp2 F F B
  FNe  :: PrimOp2 F F B

  FLt  :: PrimOp2 F F B
  FLeq :: PrimOp2 F F B
  FGt  :: PrimOp2 F F B
  FGeq :: PrimOp2 F F B

  IAdd :: PrimOp2 I I I
  ISub :: PrimOp2 I I I
  IMul :: PrimOp2 I I I
  IDiv :: PrimOp2 I I I
  IMod :: PrimOp2 I I I

  IEq  :: PrimOp2 I I B
  INe  :: PrimOp2 I I B

  ILt  :: PrimOp2 I I B
  ILeq :: PrimOp2 I I B
  IGt  :: PrimOp2 I I B
  IGeq :: PrimOp2 I I B

  And  :: PrimOp2 B B B
  Or   :: PrimOp2 B B B

-- | Constant scalars.
data SConst :: * -> * where
  SCF  :: Float -> SConst F
  SCI  :: Int -> SConst I
  SCB  :: Bool -> SConst B

------------------------------------------------------------------------
-- Show instances
------------------------------------------------------------------------
deriving instance Show (ScalarWitness a)
deriving instance Show SomeScalarWitness

instance Show (SExpr a -> SExpr b) where
  show _ = "<unary scalar function>"

instance Show (SExpr a -> SExpr b -> SExpr c) where
  show _ = "<binary scalar function>"

instance Show (SExpr a -> SExpr b -> SExpr c -> SExpr d) where
  show _ = "<ternary scalar function>"

deriving instance Show (SExpr a)
deriving instance Show (AExpr a)
deriving instance Show (MExpr a)
deriving instance Show (PrimOp1 a b)
deriving instance Show (PrimOp2 a b c)
deriving instance Show Slice
deriving instance Show MSliceInfo
deriving instance Show (SConst a)

deriving instance Typeable1 ScalarWitness
deriving instance Typeable1 SExpr
deriving instance Typeable1 AExpr
deriving instance Typeable1 MExpr
deriving instance Typeable2 PrimOp1
deriving instance Typeable3 PrimOp2
deriving instance Typeable Slice
deriving instance Typeable MSliceInfo
deriving instance Typeable1 SConst

------------------------------------------------------------------------
-- Support code for Eq instances
------------------------------------------------------------------------
-- | To test equality reliably, we need to be able to generate unique
-- names, to compare scalar functions.
type EqTest = Unique.Unique Id Bool 

sexprEq :: (Scalar a, Scalar b) => SExpr a -> SExpr b -> EqTest
sexprEq left right = maybe (return False) (sexprEq' left) (cast right)

sexprEq' :: SExpr a -> SExpr a -> EqTest
sexprEq' left right =
  case (left, right) of
    (SLet b1 f1, SLet b2 f2) -> do b <- sexprEq b1 b2
                                   f <- sexprEq (f1 b1) (f2 b2)
                                   return (b && f)
    (SVar n1, SVar n2) -> return (n1 == n2)
    (SConst c1, SConst c2) -> return (c1 == c2)
    (Prim1 op1 e1, Prim1 op2 e2) -> do
                       let op = maybe False (== op2) (cast op1)
                       e <- sexprEq e1 e2
                       return (op && e)
    (Prim2 op1 l1 r1, Prim2 op2 l2 r2) -> do
                       let op = maybe False (== op2) (cast op1)
                       l  <- sexprEq l1 l2
                       r  <- sexprEq r1 r2
                       return (op && l && r)
    (SCond p1 t1 f1, SCond p2 t2 f2) -> do
                       p <- sexprEq p1 p2
                       t <- sexprEq t1 t2
                       f <- sexprEq f1 f2
                       return (p && t && f)
    (AReduce f1 i1 t1, AReduce f2 i2 t2) -> do
      acc <- Unique.unique
      val <- Unique.unique
      f <- sexprEq (f1 (SVar acc) (SVar val)) (f2 (SVar acc) (SVar val))
      i <- sexprEq i1 i2
      t <- aexprEq t1 t2
      return (f && i && t)
    (AIdx n1 i1, AIdx n2 i2) -> do i <- sexprEq i1 i2
                                   return (n1 == n2 && i)
    (ALen a1, ALen a2) -> aexprEq a1 a2
    (MReduce f1 i1 t1, MReduce f2 i2 t2) -> do
      acc <- Unique.unique
      val <- Unique.unique
      f <- sexprEq (f1 (SVar acc) (SVar val)) (f2 (SVar acc) (SVar val))
      i <- sexprEq i1 i2
      t <- mexprEq t1 t2
      return (f && i && t)
    (MIdx n1 i1 j1, MIdx n2 i2 j2) -> do let n = n1 == n2
                                         i <- sexprEq i1 i2
                                         j <- sexprEq j1 j2
                                         return (n && i && j)
    (MRows m1, MRows m2) -> mexprEq m1 m2
    (MCols m1, MCols m2) -> mexprEq m1 m2
    _ -> return False


aexprEq :: (Scalar a, Scalar b) => AExpr a -> AExpr b -> EqTest
aexprEq left right = maybe (return False) (aexprEq' left) (cast right)

aexprEq' :: AExpr a -> AExpr a -> EqTest
aexprEq' left right =
  case (left, right) of
    (AVar n1, AVar n2) -> return (n1 == n2)
    (ASliceVar si1 n1, ASliceVar si2 n2) -> do
      si <- sliceEq si1 si2
      return (si && n1 == n2)
    (AMap f1 t1, AMap f2 t2) -> do
      val <- Unique.unique
      f <- sexprEq (f1 (SVar val)) (f2 (SVar val))
      t <- aexprEq t1 t2
      return (f && t)
    (AMap2 f1 a1 b1, AMap2 f2 a2 b2) -> do
      val1 <- Unique.unique
      val2 <- Unique.unique
      f <- sexprEq (f1 (SVar val1) (SVar val2)) (f2 (SVar val1) (SVar val2))
      a <- aexprEq a1 a2
      b <- aexprEq b1 b2
      return (f && a && b)
    (AMap3 f1 a1 b1 c1, AMap3 f2 a2 b2 c2) -> do
      val1 <- Unique.unique
      val2 <- Unique.unique
      val3 <- Unique.unique
      f <- sexprEq (f1 (SVar val1) (SVar val2) (SVar val3))
                   (f2 (SVar val1) (SVar val2) (SVar val3))
      a <- aexprEq a1 a2
      b <- aexprEq b1 b2
      c <- aexprEq c1 c2
      return (f && a && b && c)
    _ -> return False

sliceEq :: Slice -> Slice -> EqTest
sliceEq (Slice b1 e1 s1) (Slice b2 e2 s2) = do
  b <- sexprEq b1 b2
  e <- sexprEq e1 e2
  s <- sexprEq s1 s2
  return (b && e && s)

msliceInfoEq :: MSliceInfo -> MSliceInfo -> EqTest
msliceInfoEq si1 si2 = do
  rows <- sliceEq (rowSlice si1) (rowSlice si2)
  cols <- sliceEq (colSlice si1) (colSlice si2)
  return (rows && cols)

mexprEq :: (Scalar a, Scalar b) => MExpr a -> MExpr b -> EqTest
mexprEq left right = maybe (return False) (mexprEq' left) (cast right)

mexprEq' :: MExpr a -> MExpr a -> EqTest
mexprEq' left right =
  case (left, right) of
    (MVar n1, MVar n2) -> return (n1 == n2)
    (MSliceVar si1 n1, MSliceVar si2 n2) -> do
      si <- msliceInfoEq si1 si2
      return (si && n1 == n2)
    (MMap f1 t1, MMap f2 t2) -> do
      val <- Unique.unique
      f <- sexprEq (f1 (SVar val)) (f2 (SVar val))
      t <- mexprEq t1 t2
      return (f && t)
    (MMap2 f1 a1 b1, MMap2 f2 a2 b2) -> do
      val1 <- Unique.unique
      val2 <- Unique.unique
      f <- sexprEq (f1 (SVar val1) (SVar val2)) (f2 (SVar val1) (SVar val2))
      a <- mexprEq a1 a2
      b <- mexprEq b1 b2
      return (f && a && b)
    _ -> return False

------------------------------------------------------------------------
-- Eq instances
------------------------------------------------------------------------
instance Eq (ScalarWitness a) where
  WB == WB = True
  WF == WF = True
  WI == WI = True
  _  == _  = False

instance (Scalar a) => Eq (SExpr a) where
  x == y = Unique.runUnique
             (sexprEq x y)
             ["var_" ++ show i | i <- [0..] :: [Int]]

instance (Scalar a) => Eq (AExpr a) where
  x == y = Unique.runUnique
             (aexprEq x y)
             ["var_" ++ show i | i <- [0..] :: [Int]]

deriving instance Eq (SConst a)
deriving instance Eq Slice
deriving instance Eq (PrimOp1 a b)
deriving instance Eq (PrimOp2 a b c)

------------------------------------------------------------------------
-- Ord instances
------------------------------------------------------------------------
instance Eq SomeScalarWitness where
  SomeScalarWitness WB == SomeScalarWitness WB = True
  SomeScalarWitness WF == SomeScalarWitness WF = True
  SomeScalarWitness WI == SomeScalarWitness WI = True
  SomeScalarWitness _  == SomeScalarWitness _  = False

instance Ord SomeScalarWitness where
  SomeScalarWitness WB `compare` SomeScalarWitness WB = EQ
  SomeScalarWitness WB `compare` SomeScalarWitness _  = LT

  SomeScalarWitness WF `compare` SomeScalarWitness WB = GT
  SomeScalarWitness WF `compare` SomeScalarWitness WF = EQ
  SomeScalarWitness WF `compare` SomeScalarWitness WI = LT

  SomeScalarWitness WI `compare` SomeScalarWitness WB = GT
  SomeScalarWitness WI `compare` SomeScalarWitness WF = GT
  SomeScalarWitness WI `compare` SomeScalarWitness WI = EQ
