{-# LANGUAGE GADTs, ViewPatterns #-}
{-# OPTIONS -Wall #-}
module Language.Barracuda.Compiler.Aliasing
  (
    AliasInfo
  , arrayCandidates
  , matrixCandidates
  , sexprAliases
  , aexprAliases
  , mexprAliases
  , ShmemTest
  , arrayShmemTest
  , matrixShmemTest
  ) where

import Language.Barracuda.Compiler.Infrastructure
import Language.Barracuda.AST

import Control.Monad (guard)
import Control.Monad.State.Strict (StateT, execStateT, get, put, lift)
import qualified Data.Map as Map
import Data.Maybe (catMaybes)

data ArraySliceInfo
  = EntireArray
  | SliceArray ASliceInfo
  deriving (Eq, Show)

data MatrixSliceInfo
  = EntireMatrix
  | SliceMatrix MSliceInfo
  deriving (Eq, Show)

data AliasInfo = AI {
      arrayAliasInfo :: Map.Map (Id, SomeScalarWitness) [ArraySliceInfo]
    , matrixAliasInfo :: Map.Map (Id, SomeScalarWitness) [MatrixSliceInfo]
    } deriving (Eq, Show)

emptyAliasInfo :: AliasInfo
emptyAliasInfo = AI Map.empty Map.empty

arrayCandidates :: AliasInfo -> Map.Map (Id, SomeScalarWitness) Int
arrayCandidates = Map.mapMaybe dist . arrayAliasInfo
  where
    dist sis = if null ds
               then Nothing
               else if minDs < 128
               then Just minDs
               else Nothing
      where ds = [overlapDistance s1 s2 | s1 <- sis, s2 <- sis, s1 /= s2]
            minDs = minimum (catMaybes ds)

matrixCandidates :: AliasInfo -> Map.Map (Id, SomeScalarWitness) (Int, Int)
matrixCandidates = Map.mapMaybe dist . matrixAliasInfo
  where
    dist sis = if null goodDs
               then Nothing
               else Just (minimum goodDs)
      where ds = catMaybes $ do
                   s1 <- sis
                   s2 <- sis
                   guard (s1 /= s2)
                   return $ matrixOverlapDistance s1 s2
            good (r, c) = r < 8 && c < 8
            goodDs = filter good ds

type ShmemTest = (Id, SomeScalarWitness) -> Bool

arrayShmemTest :: AliasInfo -> ShmemTest
arrayShmemTest ai = \k -> Map.member k dict
  where dict = arrayCandidates ai

matrixShmemTest :: AliasInfo -> ShmemTest
matrixShmemTest ai = \k -> Map.member k dict
  where dict = matrixCandidates ai

overlapDistance :: ArraySliceInfo -> ArraySliceInfo -> Maybe Int
overlapDistance si1 si2 =
  case (si1, si2) of
    (SliceArray si1', EntireArray) -> arraySliceOffset si1'
    (EntireArray, SliceArray si2') -> arraySliceOffset si2'
    (SliceArray si1', SliceArray si2') -> do o1 <- arraySliceOffset si1'
                                             o2 <- arraySliceOffset si2'
                                             return $ abs (o1 - o2)
    (EntireArray, EntireArray) -> Nothing

arraySliceOffset :: ASliceInfo -> Maybe Int
arraySliceOffset (Slice b e _) =
  case b of
    SConst (SCI b') -> case e of
                         SConst (SCI e') -> Just (min b' e')
                         _ -> Just b'
    _ -> Nothing

matrixOverlapDistance :: MatrixSliceInfo -> MatrixSliceInfo -> Maybe (Int, Int)
matrixOverlapDistance si1 si2 =
  case (si1, si2) of
    (SliceMatrix si1', EntireMatrix) -> matrixSliceOffset si1'
    (EntireMatrix, SliceMatrix si2') -> matrixSliceOffset si2'
    (SliceMatrix si1', SliceMatrix si2') -> do
        (r1, c1) <- matrixSliceOffset si1'
        (r2, c2) <- matrixSliceOffset si2'
        return (abs $ r1 - r2, abs $ c1 - c2)
    (EntireMatrix, EntireMatrix) -> Nothing

matrixSliceOffset :: MSliceInfo -> Maybe (Int, Int)
matrixSliceOffset si = do
  ro <- arraySliceOffset (rowSlice si)
  co <- arraySliceOffset (colSlice si)
  return (ro, co)

type AliasAnalysis = StateT AliasInfo Compile ()

addArrayUse :: Id -> ScalarWitness a -> ArraySliceInfo -> AliasAnalysis
addArrayUse n w i = do
  ai <- get
  let arrInfo = arrayAliasInfo ai
  let arrInfo' = Map.insertWith' (++) (n, SomeScalarWitness w) [i] arrInfo
  put $ ai { arrayAliasInfo = arrInfo' }

addMatrixUse :: Id -> ScalarWitness a -> MatrixSliceInfo -> AliasAnalysis
addMatrixUse n w i = do
  ai <- get
  let matInfo = matrixAliasInfo ai
  let matInfo' = Map.insertWith' (++) (n, SomeScalarWitness w) [i] matInfo
  put $ ai { matrixAliasInfo = matInfo' }

sexprAliases :: (Scalar a) => SExpr a -> Compile AliasInfo
sexprAliases sexpr = execStateT (sexprAliases' sexpr) emptyAliasInfo

aexprAliases :: (Scalar a) => AExpr a -> Compile AliasInfo
aexprAliases aexpr = execStateT (aexprAliases' aexpr) emptyAliasInfo

mexprAliases :: (Scalar a) => MExpr a -> Compile AliasInfo
mexprAliases mexpr = execStateT (mexprAliases' mexpr) emptyAliasInfo


sexprAliases' :: (Scalar a) => SExpr a -> AliasAnalysis
sexprAliases' sexpr =
  case sexpr of
    SLet b f -> sexprAliases' (f b)
    SVar {} -> return ()
    SConst {} -> return ()
    Prim1 _op e -> sexprAliases' e
    Prim2 _op e1 e2 -> sexprAliases' e1 >> sexprAliases' e2
    SCond p t f -> do sexprAliases' p
                      sexprAliases' t
                      sexprAliases' f
    AReduce f i t -> do accName <- lift (uniqueId "acc")
                        valName <- lift (uniqueId "val")
                        sexprAliases' (f (SVar accName) (SVar valName))
                        sexprAliases' i
                        aexprAliases' t
    AIdx n i -> do addArrayUse n (sexprType sexpr) EntireArray
                   sexprAliases' i
    ALen {} -> return ()
    MReduce f i t -> do accName <- lift (uniqueId "acc")
                        valName <- lift (uniqueId "val")
                        sexprAliases' (f (SVar accName) (SVar valName))
                        sexprAliases' i
                        mexprAliases' t
    MIdx n i j -> do addMatrixUse n (sexprType sexpr) EntireMatrix
                     sexprAliases' i
                     sexprAliases' j
    MRows t -> mexprAliases' t
    MCols t -> mexprAliases' t
    
aexprAliases' :: (Scalar a) => AExpr a -> AliasAnalysis
aexprAliases' aexpr =
  case aexpr of
    AVar n -> addArrayUse n (aexprElemType aexpr) EntireArray
    ASliceVar si n -> do aSliceInfoAliases' si
                         addArrayUse n (aexprElemType aexpr) (SliceArray si)
    AMap f a -> do v <- lift (uniqueId "val")
                   sexprAliases' (f (SVar v))
                   aexprAliases' a
    AMap2 f a b -> do v1 <- lift (uniqueId "val")
                      v2 <- lift (uniqueId "val")
                      sexprAliases' (f (SVar v1) (SVar v2))
                      aexprAliases' a
                      aexprAliases' b
    AMap3 f a b c -> do v1 <- lift (uniqueId "val")
                        v2 <- lift (uniqueId "val")
                        v3 <- lift (uniqueId "val")
                        sexprAliases' $ f (SVar v1) (SVar v2) (SVar v3)
                        aexprAliases' a
                        aexprAliases' b
                        aexprAliases' c

mexprAliases' :: (Scalar a) => MExpr a -> AliasAnalysis
mexprAliases' mexpr =
  case mexpr of
    MVar n -> addMatrixUse n (mexprElemType mexpr) EntireMatrix
    MSliceVar si n -> do mSliceInfoAliases' si
                         addMatrixUse n (mexprElemType mexpr) (SliceMatrix si)
    MMap f a -> do v <- lift (uniqueId "val")
                   sexprAliases' (f (SVar v))
                   mexprAliases' a
    MMap2 f a b -> do v1 <- lift (uniqueId "val")
                      v2 <- lift (uniqueId "val")
                      sexprAliases' $ f (SVar v1) (SVar v2)
                      mexprAliases' a
                      mexprAliases' b

aSliceInfoAliases' :: ASliceInfo -> AliasAnalysis
aSliceInfoAliases' (Slice b e s) = do
  sexprAliases' b
  sexprAliases' e
  sexprAliases' s

mSliceInfoAliases' :: MSliceInfo -> AliasAnalysis
mSliceInfoAliases' si = do
  aSliceInfoAliases' (rowSlice si)
  aSliceInfoAliases' (colSlice si)
