{-# LANGUAGE GADTs #-}
{-# OPTIONS -Wall #-}
module Language.Barracuda.Compiler.Cpp ( compileCpp ) where

import Language.Barracuda.AST
import Language.Barracuda.SmartConstructors
  ( sliceLength
  , (!)
  , alength
  , mrows
  , mcols
  , msliceRows
  , msliceCols
  )
import Language.Barracuda.Function
import Language.Barracuda.Compiler.ArgTypes
import Language.Barracuda.Compiler.Infrastructure
import Language.Barracuda.Compiler.CppPrettyPrint
import Util.PrettyPrintHelpers

import Text.PrettyPrint

import Control.Monad (liftM, liftM3)
import Control.Monad.Writer (WriterT, tell, runWriterT, lift)

------------------------------------------------------------------------
-- Target-specific combinators for generation of pretty-printed C++
------------------------------------------------------------------------
function :: Id -> [Arg] -> Arg -> Doc -> Doc
function = \fn is o b -> funSig fn is o $+$ block b

funSig :: Id -> [Arg] -> Arg -> Doc
funSig = \fn is o -> tVoid <+> var fn <+> cpuArgList is o

------------------------------------------------------------------------
-- Compilation code
------------------------------------------------------------------------
-- | Compiles a list of functions annotated with name information into
-- C++ code.  The contents of a header file and the contents of a
-- source file are returned.
compileCpp :: CompileConfig -> [Function] -> (Doc, Doc)
compileCpp config funs = (hfile, cfile)
  where
    hfile = vsep [header, declarations, footer]
    cfile = vsep [header, functions, footer]
    header = banner "header & includes" $+$ include "barracuda.hpp"
    footer = banner "footer"
    functions = banner "functions" $+$ vsep funs'
    declarations = banner "declarations" $+$ vsep decls
    (decls, funs') = unzip $ runCompile config (mapM compileFunction funs)

-- | Compiles a Barracuda function annotated with name information.
-- Returns a prototype for the function and the compiled function
-- definition.
compileFunction :: Function -> Compile (Doc, Doc)
compileFunction (Function f n is o) = do
  let (is', o', f') = instantiate (is ++ [o]) f
  f'' <- compileSomeExpr o' f'
  return (funSig n is' o' <> semi, function n is' o' f'')

-- | Compiles an assignment statement with the given LHS identifier
-- and RHS expression.
compileSomeExpr :: Arg -> SomeExpr -> Compile Doc
compileSomeExpr outArg expr =
  case expr of
    SExpr e -> compileScalarAssignment outName e
    AExpr e -> compileArrayAssignment outName e
    MExpr e -> compileMatrixAssignment outName e
  where outName = argName outArg

-- | Compiles a scalar assignment statement.
compileScalarAssignment
  :: Id             -- ^  the name of the variable to assign
  -> SExpr a        -- ^  the scalar expression to compile
  -> Compile Doc
compileScalarAssignment outName expr = do
  (expr', assignments) <- compileSExpr expr
  return (assignments $+$ assign (var outName) expr')

compileArrayAssignment :: Id -> AExpr a -> Compile Doc
compileArrayAssignment outName expr =
  case expr of
    AVar n -> compileArrayVar outName n
    ASliceVar si n -> compileASliceVar outName si n
    AMap f a -> let rhsLen = alength a
                    rhsFun = \i -> f (a ! i)
                in compileAMap' outName rhsLen rhsFun
    AMap2 f a b -> let rhsLen = alength a
                       rhsFun = \i -> f (a ! i) (b ! i)
                   in compileAMap' outName rhsLen rhsFun
    AMap3 f a b c -> let rhsLen = alength a
                         rhsFun = \i -> f (a ! i) (b ! i) (c ! i)
                     in compileAMap' outName rhsLen rhsFun

compileMatrixAssignment :: Id -> MExpr a -> Compile Doc
compileMatrixAssignment outName expr =
  case expr of
    MVar n -> compileMatrixVar outName n
    MSliceVar si n -> compileMSliceVar outName si n
    MMap f a -> compileMMap outName f a
    MMap2 f a b -> compileMMap2 outName f a b

------------------------------------------------------------------------
-- Scalar Expression Compilation
------------------------------------------------------------------------
-- | Compile a scalar expression.  Array expressions that appear
-- within scalar expressions (such as reductions) are hoisted out and
-- replaced with variable references.  What is returned is a compiled
-- version of the scalar expression, and a C++ statement to compute
-- the hoisted assignments statements, which needs to be spliced into
-- the generated code before the compiled expression appears.
compileSExpr :: SExpr a -> Compile (Doc, Doc)
compileSExpr = runCompileScalar . compileSExpr'

-- | To accumulate the assignment statements from scalar expressions,
-- we use a WriterT monad transformer.
type CompileScalar = WriterT [Doc] Compile

runCompileScalar :: CompileScalar a -> Compile (a, Doc)
runCompileScalar = liftM (fmap vcat) . runWriterT

compileSExpr' :: SExpr a -> CompileScalar Doc
compileSExpr' expr =
  case expr of
    SLet b f      -> compileSLet b f
    SVar n        -> return (var n)
    Prim1 o e     -> compilePrim1 o e
    Prim2 o l r   -> compilePrim2 o l r
    SCond p t f   -> compileSCond p t f
    SConst c      -> return (compileSConst c)
    AReduce f i a -> compileAReduce f i a
    AIdx t i      -> compileAIdx t i
    ALen t        -> compileALen t
    MReduce f i a -> compileMReduce f i a
    MIdx t i j    -> compileMIdx t i j
    MRows t       -> compileMRows t
    MCols t       -> compileMCols t

compileSLet
  :: (Scalar a, Scalar a1)
  => SExpr a
  -> (SExpr a1 -> SExpr a2)
  -> CompileScalar Doc
compileSLet b f = do
  b' <- compileSExpr' b
  bindName <- lift $ uniqueId "let"
  let bindDecl = qConst <+> cpuScalarType (sexprType b) <+>
                 assign (var bindName) b'
  tell [bindDecl]
  compileSExpr' (f $ SVar bindName)

compilePrim1 :: PrimOp1 a b -> SExpr a -> CompileScalar Doc
compilePrim1 o e = do
  e' <- compileSExpr' e
  return $ funcall (lookupPrimOp1Function o) [e']

compilePrim2 :: PrimOp2 a b c -> SExpr a -> SExpr b -> CompileScalar Doc
compilePrim2 o l r = do
  l' <- compileSExpr' l
  r' <- compileSExpr' r
  let op = text (lookupPrimOp2Infix o)
  return (sep [ parenthesize l l', op, parenthesize r r' ])

compileSCond :: SExpr B -> SExpr a -> SExpr a -> CompileScalar Doc
compileSCond p t f = do
  p' <- compileSExpr' p
  t' <- compileSExpr' t
  f' <- compileSExpr' f
  return $ ternary (parenthesize p p')
                   (parenthesize t t')
                   (parenthesize f f')

compileSConst :: SConst a -> Doc
compileSConst c =
  case c of
    SCF f -> litFloat f
    SCI i -> int i
    SCB b -> bool b

compileAReduce
  :: (Scalar a)
  => (SExpr a -> SExpr a -> SExpr a)
  -> SExpr a
  -> AExpr a
  -> CompileScalar Doc
compileAReduce f i aexpr = do
  i' <- compileSExpr' i
  (loop, idxVarName) <- lift . loopFromZeroTo . alength $ aexpr
  accVarName <- lift $ uniqueId "reduce_acc"
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  let fInst = f (SVar accVarName) (aexpr ! SVar idxVarName)
  (f', hoisted) <- lift (compileSExpr fInst)
  let accVar = var accVarName
  let accVarDecl = cpuScalarType (sexprType i) <+> assign accVar i'
  let body = assign accVar f'
  tell [accVarDecl, loop (hoisted $+$ body)]
  return accVar

compileAIdx :: Id -> SExpr Int -> CompileScalar Doc
compileAIdx aname idx = liftM (index (var aname)) (compileSExpr' idx)

compileALen :: AExpr a -> CompileScalar Doc
compileALen expr =
  case expr of
    AVar n -> return $ method (var n) "length" []
    ASliceVar si _n -> compileSExpr' (sliceLength si)
    AMap _f a -> compileALen a
    -- TODO Check if the lengths of the two arrays are the same.
    AMap2 _f a _b -> compileALen a
    AMap3 _f a _b _c -> compileALen a

compileMReduce
  :: (Scalar a)
  => (SExpr a -> SExpr a -> SExpr a)
  -> SExpr a
  -> MExpr a
  -> CompileScalar Doc
compileMReduce f i mexpr = do
  i' <- compileSExpr' i
  (loopOuter, rowIdxName) <- lift . loopFromZeroTo . mrows $ mexpr
  (loopInner, colIdxName) <- lift . loopFromZeroTo . mcols $ mexpr
  accVarName <- lift $ uniqueId "reduce_acc"
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  let fInst = f (SVar accVarName) (mexpr ! (SVar rowIdxName, SVar colIdxName))
  (f', hoisted) <- lift (compileSExpr fInst)
  let accVar = var accVarName
  let accVarDecl = cpuScalarType (sexprType i) <+> assign accVar i'
  let body = assign accVar f'
  tell [accVarDecl, loopOuter (loopInner (hoisted $+$ body))]
  return accVar

compileMIdx :: Id -> SExpr Int -> SExpr Int -> CompileScalar Doc
compileMIdx mname rowIdx colIdx = do
  rowIdx' <- compileSExpr' rowIdx
  colIdx' <- compileSExpr' colIdx
  return $ funcall mname [rowIdx', colIdx']

compileMRows :: MExpr a -> CompileScalar Doc
compileMRows expr =
  case expr of
    MVar n -> return $ var n <> char '.' <> text "rows" <> parens empty
    MSliceVar si _n -> compileSExpr' (msliceRows si)
    MMap _f a -> compileMRows a
    MMap2 _f a _b -> compileMRows a

compileMCols :: MExpr a -> CompileScalar Doc
compileMCols expr =
  case expr of
    MVar n -> return $ var n <> char '.' <> text "cols" <> parens empty
    MSliceVar si _n -> compileSExpr' (msliceCols si)
    MMap _f a -> compileMCols a
    MMap2 _f a _b -> compileMCols a

------------------------------------------------------------------------
-- Array Expression Compilation
------------------------------------------------------------------------
compileArrayVar :: Id -> Id -> Compile Doc
compileArrayVar o v = return (funcall "array_copy" [var v, var o] <> semi)

compileASliceVar :: Id -> ASliceInfo -> Id -> Compile Doc
compileASliceVar outName sliceInfo varName = do
  ((b, e, s), hoisted) <- compileSliceInfo sliceInfo
  return $ hoisted $+$
    funcall "array_copy_slice" [var varName, b, e, s, var outName] <> semi

compileSliceInfo :: ASliceInfo -> Compile ((Doc, Doc, Doc), Doc)
compileSliceInfo (Slice b e s) =
  let cs = liftM3 (,,) (compileSExpr' b) (compileSExpr' e) (compileSExpr' s)
  in runCompileScalar cs

compileAMap'
  :: (Scalar a)
  => Id                      -- ^ output name
  -> SExpr Int               -- ^ length expression
  -> (SExpr Int -> SExpr a)  -- ^ indexing expression
  -> Compile Doc
compileAMap' outName rhsLen rhsFun = do
  (loop, idxVarName) <- loopFromZeroTo rhsLen
  (rhs', hoisted) <- compileSExpr (rhsFun $ SVar idxVarName)
  let body = assign (index (var outName) (var idxVarName)) rhs'
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  return $ loop (hoisted $+$ body)

------------------------------------------------------------------------
-- Matrix Expression Compilation
------------------------------------------------------------------------
compileMatrixVar :: Id -> Id -> Compile Doc
compileMatrixVar o v = return (funcall "matrix_copy" [var v, var o] <> semi)

compileMSliceVar :: Id -> MSliceInfo -> Id -> Compile Doc
compileMSliceVar outName sliceInfo varName = do
  (sParams, hoisted) <- compileMSliceInfo sliceInfo
  let params = var varName : sParams ++ [var outName]
  return $ hoisted $+$ funcall "matrix_copy_slice" params <> semi

compileMSliceInfo
  :: MSliceInfo -> Compile ([Doc], Doc)
compileMSliceInfo (MSliceInfo si sj) = do
  ((a, b, c), h) <- compileSliceInfo si
  ((a', b', c'), h') <- compileSliceInfo sj
  return ([a, b, c, a', b', c'], h $+$ h')

compileMMap
  :: (Scalar a)
  => Id
  -> (SExpr a -> SExpr b)
  -> MExpr a
  -> Compile Doc
compileMMap outName fun mexpr = do
  (outerLoop, outerIdxName) <- loopFromZeroTo (mrows mexpr)
  (innerLoop, innerIdxName) <- loopFromZeroTo (mcols mexpr)
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  let funInst = fun (mexpr ! (SVar outerIdxName, SVar innerIdxName))
  (fun', hoisted) <- compileSExpr funInst
  let body = assign (funcall outName [var outerIdxName, var innerIdxName]) fun'
  return $ outerLoop $ innerLoop (hoisted $+$ body)

compileMMap2
  :: (Scalar a)
  => Id
  -> (SExpr a -> SExpr b -> SExpr c)
  -> MExpr a
  -> MExpr b
  -> Compile Doc
compileMMap2 outName fun expr1 expr2 = do
  (outerLoop, outerIdxName) <- loopFromZeroTo (mrows expr1)
  (innerLoop, innerIdxName) <- loopFromZeroTo (mcols expr1)
  let outerIdxVar = SVar outerIdxName
  let innerIdxVar = SVar innerIdxName
  let idxVar = (outerIdxVar, innerIdxVar)
  -- TODO: if the hoisted portion is safe to pull out of the loop, do so.
  let funInst = fun (expr1 ! idxVar) (expr2 ! idxVar)
  (fun', hoisted) <- compileSExpr funInst
  let body = assign (funcall outName [var outerIdxName, var innerIdxName]) fun'
  return $ outerLoop $ innerLoop (hoisted $+$ body)



-- | Generates code to loop from 0 to the given stop value, exclusive.
-- Used to loop over array expressions.  Returns a function for the
-- loop body and the name of the indexing variable.
loopFromZeroTo
  :: SExpr Int                 -- ^  the stop value
  -> Compile (Doc -> Doc, Id)
loopFromZeroTo stop = do
  idxName <- uniqueId "idx"
  (stopVal, hoisted) <- compileSExpr stop
  let idxVar = var idxName
      loop body = hoisted $+$
                  forLoop (tInt <+> idxVar <+> equals <+> int 0)
                     (idxVar <+> char '<' <+> stopVal)
                     (idxVar <+> equals <+> idxVar <+> char '+' <+> int 1)
                     body
  return (loop, idxName)
