{-# LANGUAGE GADTs #-}
{-# OPTIONS -Wall #-}
-- | This module contains some pretty-printing combinators specific to
-- generating C++ fragments.
module Language.Barracuda.Compiler.CppPrettyPrint
  ( 
  -- * Pretty-Print Helpers
    var
  , refVar
  , qConst
  , qExtern
  , qStatic
  , tUnsigned
  , tFloat
  , tInt
  , tBool
  , tVoid
  , litFloat
  , block
  , forLoop
  , whileLoop
  , ifBlock
  , elseBlock
  , ifThenElse
  , ternary
  , index
  , assign
  , funcall
  , member
  , method
  , banner
  , include
  , parenthesize
  -- * Other Utility Functions
  , lookupPrimOp1Function
  , lookupPrimOp2Infix
  ) where

import Util.PrettyPrintHelpers
import Language.Barracuda.AST (SExpr(..), PrimOp1(..), PrimOp2(..))

import Text.PrettyPrint

------------------------------------------------------------------------
-- Pretty-Print Helper Functions
------------------------------------------------------------------------
var :: String -> Doc
var = text

refVar :: String -> Doc
refVar n = char '&' <> var n

qConst :: Doc
qConst = text "const"

qExtern :: Doc
qExtern = text "extern"

qStatic :: Doc
qStatic = text "static"

tUnsigned :: Doc
tUnsigned = text "unsigned"

tFloat :: Doc
tFloat = text "float"

tInt :: Doc
tInt = text "int"

tBool :: Doc
tBool = text "bool"

tVoid :: Doc
tVoid = text "void"

litFloat :: Float -> Doc
litFloat f = float f <> char 'f'

block :: Doc -> Doc
block b = char '{' $+$ nest 4 b $+$ char '}'

forLoop :: Doc -> Doc -> Doc -> Doc -> Doc
forLoop initS testS incrS body =
  text "for" <+> parens (sep [initS <> semi, testS <> semi, incrS]) $+$
  char '{' $+$
  nest 4 body $+$
  char '}'

whileLoop :: Doc -> Doc -> Doc
whileLoop test body = text "while" <+> parens test $+$ block body

ifBlock :: Doc -> Doc -> Doc
ifBlock p t = text "if" <+> parens p $+$ block t

elseBlock :: Doc -> Doc
elseBlock f = text "else" $+$ block f

ifThenElse :: Doc -> Doc -> Doc -> Doc
ifThenElse p t f = ifBlock p t $+$ elseBlock f

ternary :: Doc -> Doc -> Doc -> Doc
ternary p t f = p <+> char '?' <+> t <+> colon <+> f

index :: Doc -> Doc -> Doc
index arr idx = arr <> brackets idx

assign :: Doc -> Doc -> Doc
assign lhs rhs = lhs <+> equals <+> rhs <> semi

funcall :: String -> [Doc] -> Doc
funcall funName args = var funName <> parens (commaSep args)

member :: Doc -> String -> Doc
member d m = d <> char '.' <> text m

method :: Doc -> String -> [Doc] -> Doc
method obj methodName args = obj <> char '.' <> funcall methodName args

-- | Create a comment banner with the given text.  Useful for
-- separating sections of the generated code.
banner :: String -> Doc
banner msg = line $+$ text ("// " ++ msg) $+$ line
  where line = text ("//" ++ replicate 70 '*')

include :: String -> Doc
include file = text "#include" <+> doubleQuotes (text file)

parenthesize :: SExpr a -> Doc -> Doc
parenthesize e d = parensWhen (shouldParenthesize e) d
  where
    shouldParenthesize :: SExpr a -> Bool
    shouldParenthesize sexpr =
      case sexpr of
        SVar {}    -> False
        SConst {}  -> False
        AIdx {}    -> False
        ALen {}    -> False
        MIdx {}    -> False
        MRows {}   -> False
        MCols {}   -> False
        MReduce {} -> False
        AReduce {} -> False
        Prim1 {}   -> False
        _          -> True

lookupPrimOp1Function :: PrimOp1 a b -> String
lookupPrimOp1Function op =
  case op of
    Not    -> "not"
    INeg   -> "int_negate"
    FNeg   -> "float_negate"
    IAbs   -> "int_abs"
    FAbs   -> "fabsf"
    I2F    -> "cast_to_float"
    FSqrt  -> "sqrtf"
    FLog   -> "logf"
    FExp   -> "expf"
    FSin   -> "sinf"
    FCos   -> "cosf"
    FTan   -> "tanf"
    FASin  -> "asinf"
    FACos  -> "acosf"
    FATan  -> "atanf"
    FSinh  -> "sinhf"
    FCosh  -> "coshf"
    FTanh  -> "tanhf"
    FASinh -> "asinhf"
    FACosh -> "acoshf"
    FATanh -> "atanhf"

-- | Look up the C++ / CUDA infix operator for the given binary
-- primitive.
lookupPrimOp2Infix :: PrimOp2 a b c -> String
lookupPrimOp2Infix op =
  case op of
    FAdd -> "+"
    FSub -> "-"
    FMul -> "*"
    FDiv -> "/"
    FEq  -> "=="
    FNe  -> "!="
    FLt  -> "<"
    FLeq -> "<="
    FGt  -> ">"
    FGeq -> ">="
    IAdd -> "+"
    ISub -> "-"
    IMul -> "*"
    IDiv -> "/"
    IMod -> "%"
    IEq  -> "=="
    INe  -> "!="
    ILt  -> "<"
    ILeq -> "<="
    IGt  -> ">"
    IGeq -> ">="
    And  -> "&&"
    Or   -> "||"
