{-# LANGUAGE GADTs #-}
{-# OPTIONS -Wall #-}
module Language.Barracuda.Evaluator
  (
    evalSExpr
  , evalAExpr
  ) where

import Language.Barracuda.AST

------------------------------------------------------------------------
-- Utility Functions
------------------------------------------------------------------------
-- | Unlift a unary scalar function.
demote1 :: (Scalar a, Scalar b) => (SExpr a -> SExpr b) -> a -> b
demote1 f = evalSExpr . f . toScalar

-- | Unlift a binary scalar function.
demote2
  :: (Scalar a, Scalar b, Scalar c)
  => (SExpr a -> SExpr b -> SExpr c)
  -> a
  -> b
  -> c
demote2 f a b = evalSExpr $ f (toScalar a) (toScalar b)

-- | Unlift a ternary scalar function.
demote3
  :: (Scalar a, Scalar b, Scalar c, Scalar d)
  => (SExpr a -> SExpr b -> SExpr c -> SExpr d)
  -> a
  -> b
  -> c
  -> d
demote3 f a b c = evalSExpr $ f (toScalar a) (toScalar b) (toScalar c)

------------------------------------------------------------------------
-- Scalar Expression Evaluation
------------------------------------------------------------------------
evalSExpr :: SExpr a -> a
evalSExpr expr =
  case expr of
    SLet b f       -> let b' = toScalar (evalSExpr b) in evalSExpr (f b')
    SVar _n        -> error "evalSExpr: given an SVar"
    Prim1 o e      -> lookupPrim1 o (evalSExpr e)
    Prim2 o l r    -> lookupPrim2 o (evalSExpr l) (evalSExpr r)
    SCond p t f    -> evalSExpr (if evalSExpr p then t else f)
    SConst (SCF f) -> f
    SConst (SCI i) -> i
    SConst (SCB b) -> b
    AReduce f i a  -> foldl (demote2 f) (evalSExpr i) (evalAExpr a)
    AIdx {}        -> error "evalSExpr: given an AIdx"
    ALen t         -> length (evalAExpr t)
    MReduce f i a  -> foldl (demote2 f) (evalSExpr i) (concat $ evalMExpr a)
    MIdx {}        -> error "evalSExpr: given an MIdx"
    MRows t        -> length (evalMExpr t)
    MCols t        -> length . head . evalMExpr $ t

lookupPrim1 :: PrimOp1 a b -> a -> b
lookupPrim1 op =
  case op of
    Not    -> not
    INeg   -> negate
    FNeg   -> negate
    IAbs   -> abs
    FAbs   -> abs
    I2F    -> fromIntegral
    FSqrt  -> sqrt
    FLog   -> log
    FExp   -> exp
    FSin   -> sin
    FCos   -> cos
    FTan   -> tan
    FASin  -> asin
    FACos  -> acos
    FATan  -> atan
    FSinh  -> sinh
    FCosh  -> cosh
    FTanh  -> tanh
    FASinh -> asinh
    FACosh -> acosh
    FATanh -> atanh

lookupPrim2 :: PrimOp2 a b c -> a -> b -> c
lookupPrim2 op =
  case op of
    FAdd -> (+)
    FSub -> (-)
    FMul -> (*)
    FDiv -> (/)
    FEq  -> (==)
    FNe  -> (/=)
    FLt  -> (<)
    FLeq -> (<=)
    FGt  -> (>)
    FGeq -> (>=)
    IAdd -> (+)
    ISub -> (-)
    IMul -> (*)
    IDiv -> div
    IMod -> mod
    IEq  -> (==)
    INe  -> (/=)
    ILt  -> (<)
    ILeq -> (<=)
    IGt  -> (>)
    IGeq -> (>=)
    And  -> (&&)
    Or   -> (||)

------------------------------------------------------------------------
-- Array Expression Evaluation
------------------------------------------------------------------------
evalAExpr :: AExpr a -> [a]
evalAExpr expr =
  case expr of
    AVar _n          -> error "evalAExpr: given an AVar"
    ASliceVar _si _n -> error "evalAExpr: given an ASliceVar"
    AMap f a         -> map (demote1 f) (evalAExpr a)
    AMap2 f a b      -> zipWith (demote2 f) (evalAExpr a) (evalAExpr b)
    AMap3 f a b c    -> zipWith3 (demote3 f) (evalAExpr a) (evalAExpr b) (evalAExpr c)

------------------------------------------------------------------------
-- Matrix Expression Evaluation
------------------------------------------------------------------------
evalMExpr :: MExpr a -> [[a]]
evalMExpr expr =
  case expr of
    MVar _n -> error "evalMExpr: given an MVar"
    MSliceVar _si _n -> error "evalMExpr: given an MSliceVar"
    MMap f a -> map (map (demote1 f)) (evalMExpr a)
    MMap2 f a b -> let f' = zipWith (demote2 f)
                   in zipWith f' (evalMExpr a) (evalMExpr b)
