{-# LANGUAGE GADTs, FlexibleInstances, TypeFamilies, FlexibleContexts #-}
{-# OPTIONS -Wall -fno-warn-orphans #-}
module Language.Barracuda.SmartConstructors
  (
  -- * Smart Constructors
    int
  , float
  , bool
  , true
  , false
  , intToFloat
  , idiv
  , imod
  , (&&*)
  , (||*)
  , snot
  , ExprEq(..)
  , ExprOrd(..)
  , minVal
  , maxVal
  , ExprIndexable(..)
  , scond
  , slet
  , azipWith
  , azipWith3
  , amap
  , areduce
  , asum
  , alength
  , SliceInfo
  , aslice
  , mmap
  , mzipWith
  , mreduce
  , mrows
  , mcols
  , mslice
  -- * Convenience Functions for use in Other Modules
  , sliceLength
  , msliceRows
  , msliceCols
  , lookupOp1
  , lookupOp2
  ) where

import Language.Barracuda.AST

-- | Constant int with the given value.
int :: Int -> SExpr Int
int = SConst . SCI

-- | Constant float with the given value.
float :: Float -> SExpr Float
float = SConst . SCF

-- | Constant boolean with the given value.
bool :: Bool -> SExpr Bool
bool = SConst . SCB

-- | Constant true.
true :: SExpr Bool
true = bool True

-- | Constant false.
false :: SExpr Bool
false = bool False

-- | Cast an int to a float.
intToFloat :: SExpr Int -> SExpr Float
intToFloat v =
  case v of
    SConst (SCI i) -> float (fromIntegral i)
    _ -> Prim1 I2F v

-- | Integer division.
idiv :: SExpr Int -> SExpr Int -> SExpr Int
idiv x 1 = x
idiv x y = cfold2 div (Prim2 IDiv) x y

-- | Integer modulo.
imod :: SExpr Int -> SExpr Int -> SExpr Int
imod x 1 = x
imod x y = cfold2 mod (Prim2 IMod) x y

-- | Boolean conjunction.
(&&*) :: SExpr Bool -> SExpr Bool -> SExpr Bool
x &&* y | x == true = y
        | y == true = x
        | otherwise = cfold2 (&&) (Prim2 And) x y

-- | Boolean disjunction.
(||*) :: SExpr Bool -> SExpr Bool -> SExpr Bool
x ||* y | x == false = y
        | y == false = x
        | otherwise  = cfold2 (||) (Prim2 Or) x y

infixr 3 &&*
infixr 2 ||*

-- | Boolean negation.  This is a horrible name.
snot :: SExpr Bool -> SExpr Bool
snot b | b == false = true
       | b == true  = false
       | otherwise  = Prim1 Not b

------------------------------------------------------------------------
-- Equality Expressions
------------------------------------------------------------------------
-- | A class for expressions that can be tested for equality.  This
-- class is necessary because the built-in Eq class's functions return
-- plain Bool.
class ExprEq a where
  (==*) :: SExpr a -> SExpr a -> SExpr Bool
  (/=*) :: SExpr a -> SExpr a -> SExpr Bool

infix 4 ==*, /=*

instance ExprEq Float where
  (==*) = cfold2 (==) (Prim2 FEq)
  (/=*) = cfold2 (/=) (Prim2 FNe)

instance ExprEq Int where
  (==*) = cfold2 (==) (Prim2 IEq)
  (/=*) = cfold2 (/=) (Prim2 INe)

------------------------------------------------------------------------
-- Comparison Expressions
------------------------------------------------------------------------
-- | A class for expressions that can be compared.  This class is
-- necessary because the built-in Ord class's functions return plain
-- Bool.
class ExprOrd a where
  (>*)  :: SExpr a -> SExpr a -> SExpr Bool
  (>=*) :: SExpr a -> SExpr a -> SExpr Bool
  (<*)  :: SExpr a -> SExpr a -> SExpr Bool
  (<=*) :: SExpr a -> SExpr a -> SExpr Bool

infix 4 >*, >=*, <*, <=*

minVal :: (Scalar a, ExprOrd a) => SExpr a -> SExpr a -> SExpr a
minVal x y = slet x $ \x' ->
             slet y $ \y' ->
             scond (x' <* y') x' y'

maxVal :: (Scalar a, ExprOrd a) => SExpr a -> SExpr a -> SExpr a
maxVal x y = slet x $ \x' ->
             slet y $ \y' ->
             scond (x' >* y') x' y'

instance ExprOrd Float where
  (>*)  = cfold2 (>) (Prim2 FGt)
  (>=*) = cfold2 (>=) (Prim2 FGeq)
  (<*)  = cfold2 (<) (Prim2 FLt)
  (<=*) = cfold2 (<=) (Prim2 FLeq)

instance ExprOrd Int where
  (>*)  = cfold2 (>) (Prim2 IGt)
  (>=*) = cfold2 (>=) (Prim2 IGeq)
  (<*)  = cfold2 (<) (Prim2 ILt)
  (<=*) = cfold2 (<=) (Prim2 ILeq)

-- | Scalar if-then-else expressions.
scond :: (Scalar a) => SExpr Bool -> SExpr a -> SExpr a -> SExpr a
scond (SConst (SCB True)) t _f = t
scond (SConst (SCB False)) _t f = f
scond p t f = SCond p t f

slet :: (Scalar a, Scalar b) => SExpr a -> (SExpr a -> SExpr b) -> SExpr b
slet = SLet

-- | Map two arrays.
azipWith
  :: (Scalar a, Scalar b, Scalar c)
  => (SExpr a -> SExpr b -> SExpr c)
  -> AExpr a
  -> AExpr b
  -> AExpr c
azipWith = AMap2

-- | Map three arrays.
azipWith3
  :: (Scalar a, Scalar b, Scalar c, Scalar d)
  => (SExpr a -> SExpr b -> SExpr c -> SExpr d)
  -> AExpr a
  -> AExpr b
  -> AExpr c
  -> AExpr d
azipWith3 = AMap3

-- | Map an array.
amap
  :: (Scalar a, Scalar b)
  => (SExpr a -> SExpr b)
  -> AExpr a
  -> AExpr b
amap = AMap

-- | Reduce an array.
areduce
  :: (Scalar a)
  => (SExpr a -> SExpr a -> SExpr a)
  -> SExpr a
  -> AExpr a
  -> SExpr a
areduce = AReduce

asum :: (Num (SExpr a), Scalar a) => AExpr a -> SExpr a
asum = areduce (+) 0

class ExprIndexable a where
  type Idx a :: *
  (!) :: a t -> Idx a -> SExpr t

-- | Index an array.
aidx :: AExpr a -> SExpr Int -> SExpr a
aidx arr idx =
  case arr of
    AVar n -> AIdx n idx
    ASliceVar (Slice b _ s) n -> AVar n `aidx` (b + idx * s)
    AMap f t -> f (t `aidx` idx)
    AMap2 f t1 t2 -> slet idx $ \idx' -> f (t1 `aidx` idx') (t2 `aidx` idx')
    AMap3 f t1 t2 t3 -> slet idx $ \idx' ->
                        f (t1 `aidx` idx') (t2 `aidx` idx') (t3 `aidx` idx')

instance ExprIndexable AExpr where
  type Idx AExpr = SExpr Int
  (!) = aidx

-- | Get the length of an array.
alength :: AExpr a -> SExpr Int
alength arr =
  case arr of
    AVar _ -> ALen arr
    ASliceVar si _n -> sliceLength si
    AMap _f t -> alength t
    AMap2 _f t1 _t2 -> alength t1
    AMap3 _f t1 _t2 _t3 -> alength t1

instance ExprIndexable MExpr where
  type Idx MExpr = (SExpr Int, SExpr Int) -- row / column pair
  (!) = midx

midx :: MExpr a -> (SExpr Int, SExpr Int) -> SExpr a
midx mexpr idx@(i, j) =
  case mexpr of
    MVar n -> MIdx n i j
    MSliceVar sInfo n -> let Slice bi _ si = rowSlice sInfo
                             Slice bj _ sj = colSlice sInfo
                             i' = bi + i * si
                             j' = bj + j * sj
                         in MIdx n i' j'
    MMap f m -> f (m `midx` idx)
    MMap2 f m1 m2 -> slet i $ \i' ->
                     slet j $ \j' ->
                     let idx' = (i', j')
                     in f (m1 `midx` idx') (m2 `midx` idx')

mreduce
  :: (Scalar a)
  => (SExpr a -> SExpr a -> SExpr a)
  -> SExpr a
  -> MExpr a
  -> SExpr a
mreduce = MReduce

mmap
  :: (Scalar a, Scalar b)
  => (SExpr a -> SExpr b)
  -> MExpr a
  -> MExpr b
mmap = MMap

mzipWith
  :: (Scalar a, Scalar b, Scalar c)
  => (SExpr a -> SExpr b -> SExpr c)
  -> MExpr a
  -> MExpr b
  -> MExpr c
mzipWith = MMap2

mrows :: (Scalar a) => MExpr a -> SExpr Int
mrows = MRows

mcols :: (Scalar a) => MExpr a -> SExpr Int
mcols = MCols

type SliceInfo = (SExpr Int, SExpr Int, SExpr Int)

mslice
  :: MExpr a       -- ^  matrix to slice
  -> SliceInfo     -- ^  row slice information
  -> SliceInfo     -- ^  column slice information
  -> MExpr a
mslice expr (bi, ei, si) (bj, ej, sj) =
  let sRows = Slice bi ei si
      sCols = Slice bj ej sj
  in mslice' (MSliceInfo sRows sCols) expr

mslice' :: MSliceInfo -> MExpr a -> MExpr a
mslice' sInfo expr =
  case expr of
    MVar n -> MSliceVar sInfo n
    MSliceVar sInfo' n -> MSliceVar (mslice2 sInfo' sInfo) n
    MMap f m -> mmap f (mslice' sInfo m)
    MMap2 f m1 m2 -> mzipWith f (mslice' sInfo m1) (mslice' sInfo m2)

mslice2 :: MSliceInfo -> MSliceInfo -> MSliceInfo
mslice2 (MSliceInfo si sj) (MSliceInfo si' sj') =
    MSliceInfo (slice2 si si') (slice2 sj sj')

-- | Slice an array.
aslice
  :: AExpr a       -- ^  array expression to slice
  -> SliceInfo     -- ^  slice information
  -> AExpr a
aslice expr (start, stop, stride) = aslice' si expr
  where si = Slice start stop stride

aslice' :: Slice -> AExpr a -> AExpr a
aslice' sliceInfo expr =
  case expr of
    AVar n -> ASliceVar sliceInfo n
    ASliceVar si n -> ASliceVar (slice2 si sliceInfo) n
    AMap f a -> amap f (aslice' sliceInfo a)
    AMap2 f a b -> azipWith f (aslice' sliceInfo a) (aslice' sliceInfo b)
    AMap3 f a b c -> azipWith3
                       f
                       (aslice' sliceInfo a)
                       (aslice' sliceInfo b)
                       (aslice' sliceInfo c)

-- | Slice a slice.
slice2 :: Slice -> Slice -> Slice
slice2 (Slice b e s) (Slice b' e' s') =
  Slice (b + b' * s) (e + e' * s) s'

-- | Determines the length of a slice defined by the given parameters.
sliceLength :: Slice -> SExpr I
-- sliceLength (Slice b e s) = abs (abs (e - b) `idiv` s)
-- FIXME: do error checking for slicing.
-- This definition assumes that the stride is positive and that the
-- end index coes after the start index.
sliceLength (Slice b e s) = (e - b) `idiv` s

msliceRows :: MSliceInfo -> SExpr I
msliceRows = sliceLength . rowSlice

msliceCols :: MSliceInfo -> SExpr I
msliceCols = sliceLength . colSlice

------------------------------------------------------------
-- Num instances
------------------------------------------------------------
instance Num (SExpr Int) where
  x + 0 = x
  0 + y = y
  x + y = cfold2 (+) (Prim2 IAdd) x y
  x - 0 = x
  0 - y = - y
  x - y = cfold2 (-) (Prim2 ISub) x y
  x * 1 = x
  1 * y = y
  x * y = cfold2 (*) (Prim2 IMul) x y
  abs (Prim1 IAbs x) = Prim1 IAbs x     -- idempotency
  abs (ALen x) = ALen x                 -- array lengths always non-negative
  abs (MRows x) = MRows x               -- matrix rows always non-negative
  abs (MCols x) = MCols x               -- matrix cols always non-negative
  abs x = cfold1 abs (Prim1 IAbs) x
  fromInteger = SConst . SCI . fromInteger
  signum x = slet x $ \x' -> scond (x' <=* 0) (-1) (scond (x' ==* 0) 0 1)
  negate x = cfold1 negate (Prim1 INeg) x

instance Num (SExpr Float) where
  x + 0 = x
  0 + y = y
  x + y = cfold2 (+) (Prim2 FAdd) x y
  x - 0 = x
  0 - y = - y
  x - y = cfold2 (-) (Prim2 FSub) x y
  x * 1 = x
  1 * y = y
  x * y = cfold2 (*) (Prim2 FMul) x y
  abs (Prim1 FAbs x) = Prim1 FAbs x     -- idempotency
  abs x@(Prim1 I2F (ALen _)) = x        -- array lengths always non-negative
  abs x = cfold1 abs (Prim1 FAbs) x
  fromInteger = SConst . SCF . fromInteger
  signum x = slet x $ \x' -> scond (x' <=* 0) (-1) (scond (x' ==* 0) 0 1)
  negate x = cfold1 negate (Prim1 FNeg) x

------------------------------------------------------------
-- Fractional instances
------------------------------------------------------------
instance Fractional (SExpr Float) where
  x / 1 = x
  x / y = cfold2 (/) (Prim2 FDiv) x y
  fromRational = SConst . SCF . fromRational

------------------------------------------------------------
-- Floating instances
------------------------------------------------------------
instance Floating (SExpr Float) where
  pi    = SConst (SCF pi)
  exp   = cfold1 exp (Prim1 FExp)
  sqrt  = cfold1 sqrt (Prim1 FSqrt)
  log   = cfold1 log (Prim1 FLog)
  sin   = cfold1 sin (Prim1 FSin)
  cos   = cfold1 cos (Prim1 FCos)
  tan   = cfold1 tan (Prim1 FTan)
  asin  = cfold1 asin (Prim1 FASin)
  acos  = cfold1 acos (Prim1 FACos)
  atan  = cfold1 atan (Prim1 FATan)
  sinh  = cfold1 sinh (Prim1 FSinh)
  cosh  = cfold1 cosh (Prim1 FCosh)
  tanh  = cfold1 tanh (Prim1 FTanh)
  asinh = cfold1 asinh (Prim1 FASinh)
  acosh = cfold1 acosh (Prim1 FACosh)
  atanh = cfold1 atanh (Prim1 FATanh)

------------------------------------------------------------
-- Bounded instances
------------------------------------------------------------
instance (Scalar a, Bounded a) => Bounded (SExpr a) where
  minBound = toScalar minBound
  maxBound = toScalar maxBound

------------------------------------------------------------------------
-- Optimization Code
------------------------------------------------------------------------
-- | Generic constant folding for unary functions.  This function does
-- not recurse; it only folds the immediate expression, when possible.
cfold1
  :: (Scalar a, Scalar b)
  => (a -> b)                -- ^  the function to fold with
  -> (SExpr a -> SExpr b)    -- ^  the constructor to use otherwise
  -> SExpr a                 -- ^  the expression to inspect
  -> SExpr b
cfold1 fun _cons (SConst c) = toScalar $ fun $ fromSConst c
cfold1 _fun cons expr = cons expr

-- | Generic constant folding for binary functions.  This function
-- also does not recurse.
cfold2
  :: (Scalar a, Scalar b, Scalar c)
  => (a -> b -> c)                    -- ^  the function to fold with
  -> (SExpr a -> SExpr b -> SExpr c)  -- ^  the constructor to use otherwise
  -> SExpr a                          -- ^  the first expression to inspect
  -> SExpr b                          -- ^  the second expression to inspect
  -> SExpr c
cfold2 fun _cons (SConst c1) (SConst c2) =
  toScalar $ fun (fromSConst c1) (fromSConst c2)
cfold2 _fun cons left right = cons left right

-- | Gets the value from a scalar constant.
fromSConst :: SConst a -> a
fromSConst sconst =
  case sconst of
    SCF f -> f
    SCI i -> i
    SCB b -> b


-- | Get the smart constructor for a unary primitive op.
lookupOp1 :: PrimOp1 a b -> SExpr a -> SExpr b
lookupOp1 op1 =
  case op1 of
    Not    -> snot
    INeg   -> negate
    FNeg   -> negate
    FAbs   -> abs
    IAbs   -> abs
    I2F    -> intToFloat
    FSqrt  -> sqrt
    FLog   -> log
    FExp   -> exp
    FSin   -> sin
    FCos   -> cos
    FTan   -> tan
    FASin  -> asin
    FACos  -> acos
    FATan  -> atan
    FSinh  -> sinh
    FCosh  -> cosh
    FTanh  -> tanh
    FASinh -> asinh
    FACosh -> acosh
    FATanh -> atanh


-- | Get the smart constructor for a binary primitive op.
lookupOp2 :: PrimOp2 a b c -> SExpr a -> SExpr b -> SExpr c
lookupOp2 op2 =
  case op2 of
    FAdd -> (+)
    FSub -> (-)
    FMul -> (*)
    FDiv -> (/)
    FEq  -> (==*)
    FNe  -> (/=*)
    FLt  -> (<*)
    FLeq -> (<=*)
    FGt  -> (>*)
    FGeq -> (>=*)
    IAdd -> (+)
    ISub -> (-)
    IMul -> (*)
    IDiv -> idiv
    IMod -> imod
    IEq  -> (==*)
    INe  -> (/=*)
    ILt  -> (<*)
    ILeq -> (<=*)
    IGt  -> (>*)
    IGeq -> (>=*)
    And  -> (&&*)
    Or   -> (||*)
