#ifndef BLACK_SCHOLES_BENCHMARKS_HPP
#define BLACK_SCHOLES_BENCHMARKS_HPP


#include "array_benchmark.hpp"
#include "test_array.hpp"
#include "random.hpp"

#include "black_scholes_cuda_sdk.hpp"

#include <cublas.h>
#include <cuda_runtime.h>


class cpu_black_scholes_benchmark : public array_benchmark
{
protected:
    float_gen _generator;
    const float_array *_s;
    const float_array *_x;
    const float_array *_t;
    float_array *_result;

public:
    cpu_black_scholes_benchmark (const char * outfile_name, rand_src &rng)
        : array_benchmark(outfile_name)
        , _generator(rng, -5, 5)
        , _s(NULL)
        , _x(NULL)
        , _t(NULL)
        , _result(NULL)
        {
        }

    virtual ~cpu_black_scholes_benchmark ()
        {
        }
    
    virtual void setup (const unsigned length)
        {
            _s = new float_array (
                length,
                new_test_array(_generator, length, 10.0f, 300.0f).get()
                );
            _x = new float_array (
                length,
                new_test_array(_generator, length, 1.0f, 300.0f).get()
                );
            _t = new float_array (
                length,
                new_test_array(_generator, length, 0.25f, 10.0f).get()
                );
            _result = new float_array(length);
        }

    virtual void run ()
        {
            cpu_black_scholes(*_s, *_x, *_t, 0.02f, 0.30f, *_result);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _t;
            delete _x;
            delete _s;
        }

    virtual const char * get_name () const
        {
            return "cpu_black_scholes";
        }
};


class cuda_sdk_cpu_black_scholes_benchmark : public cpu_black_scholes_benchmark
{
public:
    cuda_sdk_cpu_black_scholes_benchmark (const char * outfile_name, rand_src &rng)
        : cpu_black_scholes_benchmark (outfile_name, rng)
        {
        }

    virtual void run ()
        {
            _run(*_s, *_x, *_t, 0.02f, 0.30f, *_result);
        }

    virtual const char * get_name () const
        {
            return "cuda_sdk_cpu_black_scholes";
        }

private:
    static inline double CND(double d){
        const double       A1 = 0.31938153;
        const double       A2 = -0.356563782;
        const double       A3 = 1.781477937;
        const double       A4 = -1.821255978;
        const double       A5 = 1.330274429;
        const double RSQRT2PI = 0.39894228040143267793994605993438;

        double
            K = 1.0 / (1.0 + 0.2316419 * fabs(d));

        double
            cnd = RSQRT2PI * exp(- 0.5 * d * d) * 
            (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

        if(d > 0)
            cnd = 1.0 - cnd;

        return cnd;
    }


    static inline void BlackScholesBodyCPU (
        float& CallResult,
        float Sf, //Stock price
        float Xf, //Option strike
        float Tf, //Option years
        float Rf, //Riskless rate
        float Vf  //Volatility rate
        )
        {
            double S = Sf, X = Xf, T = Tf, R = Rf, V = Vf;

            double sqrtT = sqrt(T);
            double    d1 = (log(S / X) + (R + 0.5 * V * V) * T) / (V * sqrtT);
            double    d2 = d1 - V * sqrtT;
            double CNDD1 = CND(d1);
            double CNDD2 = CND(d2);

            double expRT = exp(- R * T);
            CallResult = S * CNDD1 - X * expRT * CNDD2;
        }

    static inline void _run (const float_array &s,
                             const float_array &x,
                             const float_array &t,
                             const float &r,
                             const float &v,
                             float_array &result)
        {
            for(int opt = 0; opt < s.length(); opt++)
                BlackScholesBodyCPU(
                    result[opt],
                    s[opt],
                    x[opt],
                    t[opt],
                    r,
                    v
                    );
        }
};


class gpu_black_scholes_benchmark : public array_benchmark
{
protected:
    float_gen _generator;

    const gpu_float_array *_s;
    const gpu_float_array *_x;
    const gpu_float_array *_t;
    gpu_float_array *_result;

public:
    gpu_black_scholes_benchmark (const char * outfile_name, rand_src &rng)
        : array_benchmark(outfile_name)
        , _generator(rng, -5, 5)
        , _s(NULL)
        , _x(NULL)
        , _t(NULL)
        , _result(NULL)
        {
        }

    virtual ~gpu_black_scholes_benchmark ()
        {
        }
    
    virtual void setup (const unsigned length)
        {
            _s = new gpu_float_array (
                length,
                new_test_array(_generator, length, 10.0f, 300.0f).get()
                );
            _x = new gpu_float_array (
                length,
                new_test_array(_generator, length, 1.0f, 300.0f).get()
                );
            _t = new gpu_float_array (
                length,
                new_test_array(_generator, length, 0.25f, 10.0f).get()
                );
            _result = new gpu_float_array(length);
        }

    virtual void run ()
        {
            gpu_black_scholes(*_s, *_x, *_t, 0.02f, 0.30f, *_result);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _t;
            delete _x;
            delete _s;
        }

    virtual const char * get_name () const
        {
            return "gpu_black_scholes";
        }
};



class cuda_sdk_gpu_black_scholes_benchmark : public gpu_black_scholes_benchmark
{
public:
    cuda_sdk_gpu_black_scholes_benchmark (
        const char * outfile_name,
        rand_src &rng
        ) : gpu_black_scholes_benchmark (outfile_name, rng)
        {
        }

    virtual void run ()
        {
            CudaSdkBlackScholesGPU (
                _result->raw_array(),
                _s->raw_array(),
                _x->raw_array(),
                _t->raw_array(),
                0.02f,
                0.30f,
                _result->length()
                );
        }

    virtual const char * get_name () const
        {
            return "cuda_sdk_gpu_black_scholes";
        }

private:

};


#endif
