#include "black_scholes_cuda_sdk.hpp"

#include <cuda_runtime.h>


///////////////////////////////////////////////////////////////////////////////
// Polynomial approximation of cumulative normal distribution function
///////////////////////////////////////////////////////////////////////////////
static __device__ inline float cndGPU(float d){
    const float       A1 = 0.31938153f;
    const float       A2 = -0.356563782f;
    const float       A3 = 1.781477937f;
    const float       A4 = -1.821255978f;
    const float       A5 = 1.330274429f;
    const float RSQRT2PI = 0.39894228040143267793994605993438f;

    float
        K = 1.0f / (1.0f + 0.2316419f * fabsf(d));

    float
        cnd = RSQRT2PI * __expf(- 0.5f * d * d) * 
        (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

    if(d > 0)
        cnd = 1.0f - cnd;

    return cnd;
}


///////////////////////////////////////////////////////////////////////////////
// Black-Scholes formula for both call and put
///////////////////////////////////////////////////////////////////////////////
static __device__ inline void BlackScholesBodyGPU(
    float& CallResult,
    float S, //Stock price
    float X, //Option strike
    float T, //Option years
    float R, //Riskless rate
    float V  //Volatility rate
    ){
    float sqrtT, expRT;
    float d1, d2, CNDD1, CNDD2;

    sqrtT = sqrtf(T);
    d1 = (__logf(S / X) + (R + 0.5f * V * V) * T) / (V * sqrtT);
    d2 = d1 - V * sqrtT;

    CNDD1 = cndGPU(d1);
    CNDD2 = cndGPU(d2);

    expRT = __expf(- R * T);
    CallResult = S * CNDD1 - X * expRT * CNDD2;
}


////////////////////////////////////////////////////////////////////////////////
//Process an array of optN options on GPU
////////////////////////////////////////////////////////////////////////////////
__global__ void BlackScholesGPU(
    float *d_CallResult,
    const float *d_StockPrice,
    const float *d_OptionStrike,
    const float *d_OptionYears,
    float Riskfree,
    float Volatility,
    int optN
    ){
    //Thread index
    const int      tid = blockDim.x * blockIdx.x + threadIdx.x;
    //Total number of threads in execution grid
    const int THREAD_N = blockDim.x * gridDim.x;

    //No matter how small is execution grid or how large OptN is,
    //exactly OptN indices will be processed with perfect memory coalescing
    for(int opt = tid; opt < optN; opt += THREAD_N)
        BlackScholesBodyGPU(
            d_CallResult[opt],
            d_StockPrice[opt],
            d_OptionStrike[opt],
            d_OptionYears[opt],
            Riskfree,
            Volatility
            );
}



void CudaSdkBlackScholesGPU(
    float *d_CallResult,
    const float *d_StockPrice,
    const float *d_OptionStrike,
    const float *d_OptionYears,
    float Riskfree,
    float Volatility,
    int optN
    )
{
    const unsigned block_dims = 256;
    const unsigned grid_dims = (optN + (block_dims - 1)) / block_dims;
    BlackScholesGPU<<<grid_dims, block_dims>>> (
        d_CallResult,
        d_StockPrice,
        d_OptionStrike,
        d_OptionYears,
        Riskfree,
        Volatility,
        optN
        );
}
