#include "barracuda.hpp"
#include <cuda_runtime.h>


static __global__
void
jacobi_stencil (
    kernel_float_matrix in,
    kernel_float_matrix_view out
    )
{
    const int tx = threadIdx.x;
    const int ty = threadIdx.y;

    const int x = blockIdx.x * 16 + tx;
    const int y = blockIdx.y * 16 + ty;

    // Block-specific shared memory.  There is some overlap among
    // blocks, so that all cells can compute the stencil at all their
    // points.
    __shared__ float smem[16][16];

    // Load from GPU memory into shared memory.
    if (x < in.cols() && y < in.rows())
        smem[ty][tx] = in(y, x);
    __syncthreads();

    // Compute the stencil, writing directly to output array in GPU
    // memory.  The border cases are handled specially.
    if (0 < x && x < in.cols()-1 && 0 < y && y < in.rows()-1){
        float sum;

        sum = (tx == 0)
            ? in(y, x-1)
            : smem[ty][tx-1];

        sum += (tx == 32 - 1)
            ?  in(y, x+1)
            :  smem[ty][tx+1];

        sum += (ty == 0)
            ?  in(y-1, x)
            :  smem[ty-1][tx];

        sum += (ty == 32 - 1)
            ?  in(y+1, x)
            :  smem[ty+1][tx];

        out(y, x) = sum / 4.0f;
    }
}


void
handwritten_jacobi_stencil_shmem (
    const gpu_float_matrix &input,
    gpu_float_matrix_view &output
    )
{
    const dim3 block_dims (16, 16, 1);
    const dim3 grid_dims (
        (input.cols() + 15) / 16,
        (input.rows() + 15) / 16,
        1
        );
    jacobi_stencil<<<grid_dims, block_dims>>> (
        kernel_float_matrix (input),
        kernel_float_matrix_view (output)
        );
}
