#ifndef JACOBI_STENCIL_BENCHMARKS_HPP
#define JACOBI_STENCIL_BENCHMARKS_HPP


#include "matrix_benchmark.hpp"
#include "test_array.hpp"
#include "random.hpp"

#include <cublas.h>
#include <cuda_runtime.h>

class cpu_jacobi_stencil_benchmark : public matrix_benchmark
{
protected:
    float_gen _generator;
    const float_matrix *_input;
    float_matrix *_result;

public:
    cpu_jacobi_stencil_benchmark (const char * outfile_name, rand_src &rng)
        : matrix_benchmark(outfile_name)
        , _generator(rng, 20.0f, 25.0f)
        , _input(NULL)
        , _result(NULL)
        {
        }

    virtual ~cpu_jacobi_stencil_benchmark ()
        {
        }
    
    virtual void setup (const unsigned rows, const unsigned cols)
        {
            _input = new float_matrix (
                rows,
                cols,
                new_test_array(_generator, rows * cols).get()
                );
            _result = new float_matrix(rows, cols);
        }

    virtual void run ()
        {
            float_matrix_view result_view (
                *_result,
                slice(1, _result->rows() - 1, 1),
                slice(1, _result->cols() - 1, 1)
                );
            cpu_jacobi_stencil(*_input, result_view);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _input;
        }

    virtual const char * get_name () const
        {
            return "cpu_jacobi_stencil";
        }
};


class handwritten_cpu_jacobi_stencil_benchmark : public cpu_jacobi_stencil_benchmark
{
public:
    handwritten_cpu_jacobi_stencil_benchmark (const char * outfile_name, rand_src &rng)
        : cpu_jacobi_stencil_benchmark (outfile_name, rng)
        {
        }

    virtual void run ()
        {
            for (int y = 1; y < _input->rows() - 1; y += 1)
                for (int x = 1; x < _input->cols() - 1; x += 1)
                    (*_result)(y, x) =
                        ((*_input)(y-1, x) +
                         (*_input)(y+1, x) +
                         (*_input)(y, x-1) +
                         (*_input)(y, x+1)) / 4;
        }

    virtual const char * get_name () const
        {
            return "handwritten_cpu_jacobi_stencil";
        }
};


class gpu_noshmem_jacobi_stencil_benchmark : public matrix_benchmark
{
protected:
    float_gen _generator;

    const gpu_float_matrix *_input;
    gpu_float_matrix *_result;

public:
    gpu_noshmem_jacobi_stencil_benchmark (
        const char * outfile_name,
        rand_src &rng
        ) : matrix_benchmark(outfile_name)
          , _generator(rng, 20.0f, 25.0f)
          , _input(NULL)
          , _result(NULL)
        {
        }

    virtual ~gpu_noshmem_jacobi_stencil_benchmark ()
        {
        }
    
    virtual void setup (const unsigned rows, const unsigned cols)
        {
            _input = new gpu_float_matrix (
                rows,
                cols,
                new_test_array(_generator, rows * cols).get()
                );
            _result = new gpu_float_matrix(rows, cols);
        }

    virtual void run ()
        {
            gpu_float_matrix_view result_view (
                *_result,
                slice (1, _result->rows() - 1, 1),
                slice (1, _result->cols() - 1, 1)
                );
            gpu_noshmem_jacobi_stencil(*_input, result_view);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _input;
        }

    virtual const char * get_name () const
        {
            return "gpu_noshmem_jacobi_stencil";
        }
};


class gpu_shmem_jacobi_stencil_benchmark : public matrix_benchmark
{
protected:
    float_gen _generator;

    const gpu_float_matrix *_input;
    gpu_float_matrix *_result;

public:
    gpu_shmem_jacobi_stencil_benchmark (
        const char * outfile_name,
        rand_src &rng
        ) : matrix_benchmark(outfile_name)
          , _generator(rng, 20.0f, 25.0f)
          , _input(NULL)
          , _result(NULL)
        {
        }

    virtual ~gpu_shmem_jacobi_stencil_benchmark ()
        {
        }
    
    virtual void setup (const unsigned rows, const unsigned cols)
        {
            _input = new gpu_float_matrix (
                rows,
                cols,
                new_test_array(_generator, rows * cols).get()
                );
            _result = new gpu_float_matrix(rows, cols);
        }

    virtual void run ()
        {
            gpu_float_matrix_view result_view (
                *_result,
                slice (1, _result->rows() - 1, 1),
                slice (1, _result->cols() - 1, 1)
                );
            gpu_shmem_jacobi_stencil(*_input, result_view);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _input;
        }

    virtual const char * get_name () const
        {
            return "gpu_shmem_jacobi_stencil";
        }
};


void handwritten_jacobi_stencil_noshmem (
    const gpu_float_matrix &input,
    gpu_float_matrix_view &output
    );

class handwritten_gpu_noshmem_jacobi_stencil_benchmark
    : public gpu_noshmem_jacobi_stencil_benchmark
{
public:
    handwritten_gpu_noshmem_jacobi_stencil_benchmark (
        const char * outfile_name,
        rand_src &rng
        ) : gpu_noshmem_jacobi_stencil_benchmark(outfile_name, rng)
        {
        }

    virtual ~handwritten_gpu_noshmem_jacobi_stencil_benchmark ()
        {
        }
    
    virtual void run ()
        {
            gpu_float_matrix_view result_view (
                *_result,
                slice (1, _result->rows() - 1, 1),
                slice (1, _result->cols() - 1, 1)
                );
            handwritten_jacobi_stencil_noshmem (*_input, result_view);
        }

    virtual const char * get_name () const
        {
            return "handwritten_gpu_jacobi_stencil_noshmem";
        }
};


void handwritten_jacobi_stencil_shmem (
    const gpu_float_matrix &input,
    gpu_float_matrix_view &output
    );

class handwritten_gpu_shmem_jacobi_stencil_benchmark
    : public gpu_shmem_jacobi_stencil_benchmark
{
public:
    handwritten_gpu_shmem_jacobi_stencil_benchmark (
        const char * outfile_name,
        rand_src &rng
        ) : gpu_shmem_jacobi_stencil_benchmark(outfile_name, rng)
        {
        }

    virtual ~handwritten_gpu_shmem_jacobi_stencil_benchmark ()
        {
        }
    
    virtual void run ()
        {
            gpu_float_matrix_view result_view (
                *_result,
                slice (1, _result->rows() - 1, 1),
                slice (1, _result->cols() - 1, 1)
                );
            handwritten_jacobi_stencil_shmem (*_input, result_view);
        }

    virtual const char * get_name () const
        {
            return "handwritten_gpu_jacobi_stencil_shmem";
        }
};


#endif
