#ifndef SAXPY_BENCHMARKS_HPP
#define SAXPY_BENCHMARKS_HPP


#include "array_benchmark.hpp"
#include "test_array.hpp"
#include "random.hpp"

#include <cublas.h>
#include <cuda_runtime.h>


class cpu_saxpy_benchmark : public array_benchmark
{
private:
    float_gen _generator;
    const float *_alpha;
    const float_array *_x;
    const float_array *_y;
    float_array *_result;

public:
    cpu_saxpy_benchmark (const char * outfile_name, rand_src &rng)
        : array_benchmark(outfile_name)
        , _generator(rng, -5, 5)
        , _alpha(NULL)
        , _x(NULL)
        , _y(NULL)
        , _result(NULL)
        {
        }

    virtual ~cpu_saxpy_benchmark ()
        {
        }
    
    virtual void setup (const unsigned length)
        {
            _alpha = new float(_generator());
            _x = new float_array (
                length,
                new_test_array(_generator, length).get()
                );
            _y = new float_array (
                length,
                new_test_array(_generator, length).get()
                );
            _result = new float_array(length);
        }

    virtual void run ()
        {
            cpu_saxpy(*_alpha, *_x, *_y, *_result);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _y;
            delete _x;
            delete _alpha;
        }

    virtual const char * get_name () const
        {
            return "cpu_saxpy";
        }
};


class gpu_saxpy_benchmark : public array_benchmark
{
private:
    float_gen _generator;

    const float *_alpha;
    const gpu_float_array *_x;
    const gpu_float_array *_y;
    gpu_float_array *_result;

public:
    gpu_saxpy_benchmark (const char * outfile_name, rand_src &rng)
        : array_benchmark(outfile_name)
        , _generator(rng, -5, 5)
        , _alpha(NULL)
        , _x(NULL)
        , _y(NULL)
        , _result(NULL)
        {
        }

    virtual ~gpu_saxpy_benchmark ()
        {
        }
    
    virtual void setup (const unsigned length)
        {
            _alpha = new float(_generator());
            _x = new gpu_float_array (
                length,
                new_test_array(_generator, length).get()
                );
            _y = new gpu_float_array (
                length,
                new_test_array(_generator, length).get()
                );
            _result = new gpu_float_array(length);
        }

    virtual void run ()
        {
            gpu_saxpy(*_alpha, *_x, *_y, *_result);
        }

    virtual void teardown ()
        {
            delete _result;
            delete _y;
            delete _x;
            delete _alpha;
        }

    virtual const char * get_name () const
        {
            return "gpu_saxpy";
        }
};


class cublas_saxpy_benchmark : public array_benchmark
{
private:
    float_gen _generator;

    const float *_alpha;
    const gpu_float_array *_x;
    gpu_float_array *_y;

public:
    cublas_saxpy_benchmark (const char * outfile_name, rand_src &rng)
        : array_benchmark(outfile_name)
        , _generator(rng, -5, 5)
        , _alpha(NULL)
        , _x(NULL)
        , _y(NULL)
        {
            cublasInit();
        }

    virtual ~cublas_saxpy_benchmark ()
        {
            cublasShutdown();
        }
    
    virtual void setup (const unsigned length)
        {
            _alpha = new float(_generator());
            _x = new gpu_float_array(
                length,
                new_test_array(_generator, length).get()
                );
            _y = new gpu_float_array(
                length,
                new_test_array(_generator, length).get()
                );
        }

    virtual void run ()
        {
            cublasSaxpy (
                _x->length(),
                *_alpha,
                _x->raw_array(),
                1,
                _y->raw_array(),
                1
                );
        }

    virtual void teardown ()
        {
            delete _y;
            delete _x;
            delete _alpha;
        }

    virtual const char * get_name () const
        {
            return "cublas_saxpy";
        }
};


#endif
