import Language.Barracuda
import Data.Maybe (mapMaybe)

-- | Really simple:  scalar addition.
scalarPlus :: SExpr Float -> SExpr Float -> SExpr Float
scalarPlus = (+)

-- | Calculates the length of hypotenuse of a right triangle given the
-- other two sides.  An application of Pythagorean's Theorem.
hypotenuse :: SExpr Float -> SExpr Float -> SExpr Float
hypotenuse a b = sqrt (a^2 + b^2)

-- | Add one to each element of an array.
plusOne :: AExpr Float -> AExpr Float
plusOne arr = amap (+ 1) arr

-- | Add the sum of the first array to each element of the second.
-- Demonstrates nested use of the array operations.
addSum :: AExpr Float -> AExpr Float -> AExpr Float
addSum arr1 arr2 = amap (asum arr1 +) arr2

-- | Add the sum of the first array to each element of the second, and
-- compute the sum of the resulting array.
sumAddSum :: AExpr Float -> AExpr Float -> SExpr Float
sumAddSum arr1 arr2 = asum (addSum arr1 arr2)

-- | SAXPY from BLAS Level 1.
saxpy :: SExpr Float -> AExpr Float -> AExpr Float -> AExpr Float
saxpy a x y = azipWith (+) (amap (* a) x) y

-- | SDOT from BLAS Level 1.
sdot :: AExpr Float -> AExpr Float -> SExpr Float
sdot x y = asum $ azipWith (*) x y

-- | The following definition uses nested data parallelism that
-- cannot be simply hoisted out.
ndp :: AExpr Float -> AExpr Float -> AExpr Float
ndp ndpArr = amap (asum . f)
  where f v = amap (+ v) ndpArr

-- | Root Mean Square Error.
rmse :: AExpr Float -> AExpr Float -> SExpr Float
rmse x y = sqrt (sumDiff / len)
  where
    len = intToFloat (alength x)
    sumDiff = asum $ amap (^2) $ azipWith (-) x y

-- | Copy from a slice of the given array.
copySlice :: AExpr Float -> AExpr Float
copySlice a1 = aslice a1 (10, 100, 1)

-- | Scalar maximum.
smax x y = scond (x >=* y) x y

-- | Take the maximum of two arrays.
amax :: AExpr Float -> AExpr Float -> AExpr Float
amax = azipWith smax

-- | Forward difference of an array.
forwardDiff :: AExpr Float -> AExpr Float
forwardDiff x = azipWith (-) x'' x'
  where x' = aslice x (0, alength x - 1, 1)
        x'' = aslice x (1, alength x, 1)

-- | SAXPY on two disjoint array slices from the same array.
saxpySlices :: SExpr Float -> AExpr Float -> AExpr Float
saxpySlices alpha arr = saxpy alpha x y
  where x = aslice arr (10, 20, 1)
        y = aslice arr (40, 50, 1)

-- | SAXPY on array slices from two arrays.
saxpySlices2 :: SExpr Float -> AExpr Float -> AExpr Float -> AExpr Float
saxpySlices2 alpha arr1 arr2 = saxpy alpha x y
  where x = aslice arr1 (10, 20, 1)
        y = aslice arr2 (40, 50, 1)

backpermute :: AExpr Float -> AExpr Int -> AExpr Float
backpermute arr idxs = amap (arr !) idxs

-- | One step of Jacobi iteration
jacobiStep :: MExpr Float -> MExpr Float
jacobiStep m = mmap (/ 4) (n .+. s .+. e .+. w)
  where
    (.+.) = mzipWith (+)
    n = mslice m (0, mrows m - 2, 1) (1, mcols m - 1, 1)
    s = mslice m (2, mrows m, 1)     (1, mcols m - 1, 1)
    e = mslice m (1, mrows m - 1, 1) (2, mcols m, 1)
    w = mslice m (1, mrows m - 1, 1) (0, mcols m - 2, 1)

-- | The maximum difference computation, to determine convergence.
jacobiDiff :: MExpr Float -> MExpr Float -> SExpr Float
jacobiDiff a1 a2 = mreduce maxVal (-1e12) diffArr
  where
    diffArr = mzipWith (\x y -> abs (x - y)) a1 a2
    maxVal x y = scond (x >=* y) x y

jacobiTest :: [Function]
jacobiTest =
  [ Function jacobiStep "jacobi_iterate" ["input"] "result"
  , Function jacobiDiff "jacobi_difference" ["array1", "array2"] "result"
  ]

rmseTest :: [Function]
rmseTest = [Function rmse "rmse" ["input1", "input2"] "result"]

-- | A compilation unit, i.e. a list of functions on expressions,
-- annotated with name information.  This can be given to the
-- `compileCpp' or `compileCuda' functions.
compileUnit :: [Function]
compileUnit =
  [ Function scalarPlus "scalarPlus" ["x", "y"] "result"
  , Function hypotenuse "hypotenuse" ["a", "b"] "result"
  , Function plusOne "plusOne" ["arr"] "result"
  , Function (asum :: AExpr Float -> SExpr Float) "asum" ["arr"] "result"
  , Function addSum "addSum" ["arr1", "arr2"] "result"
  , Function sumAddSum "sumAddSum" ["arr1", "arr2"] "result"
  , Function saxpy "saxpy" ["alpha", "x", "y"] "result"
  , Function sdot "sdot" ["x", "y"] "result"
  , Function ndp "ndp" ["arr1", "arr2"] "result"
  , Function rmse "rmse" ["x", "y"] "result"
  , Function amax "amax" ["x", "y"] "result"
  , Function forwardDiff "forwardDiff" ["x"] "result"
  , Function saxpySlices "saxpySlices" ["alpha", "array"] "result"
  , Function saxpySlices2 "saxpySlices2" ["alpha", "x", "y"] "result"
  , Function backpermute "backpermute" ["values", "indexes"] "result"
  ]

cudaTests :: [Function]
cudaTests = mapMaybe lookupByName
  [ "scalarPlus"
  , "hypotenuse"
  , "plusOne"
  , "saxpy"
  , "ndp2"
  ]

-- | Lookup a function entry in the compilation unit by name.
lookupByName :: String -> Maybe Function
lookupByName n = lookup n funs
  where funs = [ (functionName f, f) | f <- compileUnit ]
